package de.drmaxnix.mfcbalance

import android.app.PendingIntent
import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.nfc.NfcAdapter
import android.nfc.Tag
import android.nfc.tech.MifareClassic
import android.os.Build
import android.os.Bundle
import androidx.activity.ComponentActivity
import androidx.activity.compose.BackHandler
import androidx.activity.compose.setContent
import androidx.activity.enableEdgeToEdge
import androidx.compose.animation.fadeIn
import androidx.compose.animation.fadeOut
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.padding
import androidx.compose.material3.Scaffold
import androidx.compose.runtime.Composable
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableIntStateOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import androidx.navigation.compose.NavHost
import androidx.navigation.compose.composable
import androidx.navigation.compose.rememberNavController
import androidx.navigation.createGraph
import androidx.preference.PreferenceManager
import de.drmaxnix.mfcbalance.screen.Home
import de.drmaxnix.mfcbalance.screen.Settings
import de.drmaxnix.mfcbalance.ui.theme.MFCBalanceTheme
import de.drmaxnix.mfcbalance.util.BalanceFormatter
import de.drmaxnix.mfcbalance.util.DebugLog
import de.drmaxnix.mfcbalance.util.MfcReader
import de.drmaxnix.mfcbalance.util.MfcReader.Companion.byteArrayToHexString
import de.drmaxnix.mfcbalance.util.MfcReader.Companion.hexStringToByteArray
import kotlinx.coroutines.delay
import kotlinx.coroutines.flow.flow
import kotlin.math.floor

class MainActivity : ComponentActivity() {
    private var nfcAdapter: NfcAdapter? = null
    private var scannedAt: Long = 0

    private var mutableInstructions: String = ""; private var mutableInstructionsRef = mutableStateOf(mutableInstructions)
    private var mutableShowNfcSettingsButton: Boolean = false; private var mutableShowNfcSettingsButtonRef = mutableStateOf(mutableShowNfcSettingsButton)
    private var mutableBalance: String = ""; private var mutableBalanceRef = mutableStateOf(mutableBalance)

    override fun onCreate(savedInstanceState: Bundle?){
        super.onCreate(savedInstanceState)

        nfcAdapter = NfcAdapter.getDefaultAdapter(this)
        mutableReset()
        this.registerReceiver(mNfcStateChangedReceiver, IntentFilter(NfcAdapter.ACTION_ADAPTER_STATE_CHANGED))
        updateNfcEnabledState()

        enableEdgeToEdge()
        setContent {
            MFCBalanceTheme {
                MainScreen()
            }
        }
    }

    override fun onResume() {
        super.onResume()
        enableNfcForegroundDispatch()
    }

    override fun onPause() {
        super.onPause()
        disableNfcForegroundDispatch()
    }

    override fun onDestroy() {
        super.onDestroy()
        this.unregisterReceiver(mNfcStateChangedReceiver);
    }

    private fun mutableReset(){
        DebugLog.clear()
        mutableInstructionsRef.value = "Tap a tag or card to see its balance"
        mutableShowNfcSettingsButtonRef.value = false
        mutableBalanceRef.value = ""
        scannedAt = 0
    }

    private fun updateNfcEnabledState(){
        // SHOW INSTRUCTIONS IF NOT ENABLED //
        if(nfcAdapter?.isEnabled != true){
            // not enabled, clear previous results, show instructions
            mutableInstructionsRef.value = "NFC is disabled on your device"
            mutableShowNfcSettingsButtonRef.value = true
            mutableBalanceRef.value = ""
            return
        }


        // WAS DISABLED PREVIOUSLY? //
        if(mutableShowNfcSettingsButtonRef.value){
            // reset balance card
            mutableReset()

            // listen for nfc events
            enableNfcForegroundDispatch()
        }
    }

    override fun onNewIntent(intent: Intent?) {
        super.onNewIntent(intent)
        intent?.let {
            if(intent.action.equals(NfcAdapter.ACTION_TAG_DISCOVERED)){
                mutableReset()
                DebugLog.addLine(level = 0, line = "NFC intent received")

                // GET TAG //
                DebugLog.addLine(level = 0, line = "Getting tag object from intent")

                // obtain from intent
                DebugLog.addLine(level = 1, line = "SDK version: " + Build.VERSION.SDK_INT.toString())
                val tag = if(Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU){
                    intent.getParcelableExtra(NfcAdapter.EXTRA_TAG, Tag::class.java)
                } else {
                    @Suppress("DEPRECATION")
                    intent.getParcelableExtra(NfcAdapter.EXTRA_TAG)
                }

                // make sure it is not null
                if(tag == null){
                    DebugLog.addLine(level = 1, line = "Tag object is null; aborting")
                    mutableInstructionsRef.value = "Incompatible Tag"
                    return
                }
                DebugLog.addLine(level = 1, line = "Tag object obtained")
                DebugLog.addLine(level = 2, line = "Tag id: " + MfcReader.byteArrayToHexString(tag.id))
                DebugLog.addLine(level = 2, line = "Tag techs: " + tag.techList.joinToString())

                // make sure it is a classic tag
                if(!tag.techList.contains(MifareClassic::class.java.name)){
                    DebugLog.addLine(level = 1, line = "Not a Mifare Classic tag; aborting")
                    mutableInstructionsRef.value = "Incompatible Tag"
                    return
                }


                // PREPARE SETTINGS //
                DebugLog.addLine(level = 0, line = "Preparing settings")
                val preferenceManager = PreferenceManager.getDefaultSharedPreferences(this.applicationContext)

                // sector index
                val sector = preferenceManager.getInt("mfccard_sector", 0)
                DebugLog.addLine(level = 1, line = "Sector: $sector")

                // sector key
                val keyString = preferenceManager.getString("mfccard_key", "ffffffffffff")!!
                val keyBytes = hexStringToByteArray(keyString)
                DebugLog.addLine(level = 1, line = "Sector key: ${byteArrayToHexString(keyBytes)}")

                // block in sector
                val blockInSector = preferenceManager.getInt("mfccard_block_in_sector", 0)
                DebugLog.addLine(level = 1, line = "Block in Sector: $blockInSector")

                // byte offset
                val byteOffset = preferenceManager.getInt("mfccard_byte_offset_in_block", 0)
                DebugLog.addLine(level = 1, line = "Byte offset: $byteOffset")

                // number size
                val numberSize = preferenceManager.getInt("mfccard_number_size", 16)
                DebugLog.addLine(level = 1, line = "Number size: $numberSize bit")

                // byte order
                val byteorder = preferenceManager.getString("mfccard_byteorder", "be")!!
                DebugLog.addLine(level = 1, line = "Byte order: $byteorder")

                // divisor
                val divisor = preferenceManager.getInt("balancedisplay_divisor", 100)
                DebugLog.addLine(level = 1, line = "Divisor: $divisor")

                // currency symbol
                val currencySymbol = preferenceManager.getString("balancedisplay_currency_symbol", "$")!!
                DebugLog.addLine(level = 1, line = "Currency symbol: $currencySymbol")


                // READ NUMBER FROM TAG //
                val reader = MfcReader(tag = tag)
                val number = reader.getNumber(
                    sector = sector,
                    keyBytes = keyBytes,
                    blockInSector = blockInSector,
                    byteOffset = byteOffset,
                    numberSize = numberSize,
                    byteorder = byteorder
                )
                if(number == null){
                    // failed to read, display instructions
                    mutableInstructionsRef.value = reader.getInstructions()
                    return
                }


                // DISPLAY NUMBER AS BALANCE //
                // get with formatting
                val balance = BalanceFormatter.numberFormat(
                    number = number,
                    divisor = divisor,
                    currencySymbol = currencySymbol
                )

                // reset age timer
                scannedAt = (System.currentTimeMillis() / 1000)

                // set display text
                mutableBalanceRef.value = balance
            }
        }
    }

    private fun enableNfcForegroundDispatch() {
        nfcAdapter?.let { adapter ->
            if(!adapter.isEnabled) return

            val pendingIntent = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.S) {
                PendingIntent.getActivity(
                    this,
                    0,
                    Intent(this, javaClass).addFlags(Intent.FLAG_ACTIVITY_SINGLE_TOP),
                    PendingIntent.FLAG_MUTABLE or PendingIntent.FLAG_CANCEL_CURRENT
                )
            } else {
                PendingIntent.getActivity(
                    this,
                    0,
                    Intent(this, javaClass).addFlags(Intent.FLAG_ACTIVITY_SINGLE_TOP),
                    PendingIntent.FLAG_UPDATE_CURRENT
                )
            }

            adapter.enableForegroundDispatch(
                this, pendingIntent, arrayOf(
                    IntentFilter(NfcAdapter.ACTION_TAG_DISCOVERED)
                ), null
            )
        }
    }

    private fun disableNfcForegroundDispatch() {
        nfcAdapter?.disableForegroundDispatch(this)
    }

    private fun ageStringFormat(): String {
        if(scannedAt <= 0) return "now"

        val now = (System.currentTimeMillis() / 1000)
        val age = now - scannedAt

        if(age <= 5) return "now"
        if(age <= 60) return "a few seconds ago"
        if(age <= 120) return "a minute ago"

        val minuteCount = floor(age.toDouble() / 60.0).toInt()
        if(minuteCount < 60) return "$minuteCount minutes ago"

        val hourCount = floor(minuteCount.toDouble() / 60.0).toInt()
        if(hourCount <= 1) return "an hour ago"
        if(hourCount <= 20) return "$hourCount hours ago"

        val dayCount = floor(hourCount.toDouble() / 24.0).toInt()
        if(dayCount <= 1) return "a day ago"
        return "$dayCount days ago"
    }

    @Composable
    fun MainScreen(){
        val navController = rememberNavController()
        val selectedNavigationIndex = rememberSaveable { mutableIntStateOf(0) }

        val ageStringRef = flow {
            while(true){
                emit(ageStringFormat())
                delay(100)
            }
        }.collectAsState(ageStringFormat())

        Scaffold(
            modifier = Modifier.fillMaxSize(),
            bottomBar = { BottomNavigationBar(
                navController = navController,
                selectedNavigationIndex = selectedNavigationIndex,
            ) },
        ){ innerPadding ->
            val graph = navController.createGraph(startDestination = Screen.Home.route){
                composable(route = Screen.Home.route){
                    val instructions by mutableInstructionsRef
                    val showNfcSettingsButton by mutableShowNfcSettingsButtonRef
                    val balance by mutableBalanceRef
                    val debugText by DebugLog.getMutableDebugTextRef()
                    Home(
                        instructions = instructions,
                        showNfcSettingsButton = showNfcSettingsButton,
                        balance = balance,
                        ageString = ageStringRef.value,
                        debugText = debugText
                    )
                }
                composable(route = Screen.Settings.route) {
                    Settings()
                }
            }

            NavHost(
                navController = navController,
                graph = graph,
                modifier = Modifier.padding(innerPadding),
                enterTransition = { fadeIn() },
                exitTransition = { fadeOut() },
            )

            BackHandler {
                if(navController.currentDestination?.route == Screen.Home.route)
                    finishAffinity()

                selectedNavigationIndex.intValue = 0
                navController.navigate(Screen.Home.route)
            }
        }
    }

    private val mNfcStateChangedReceiver: BroadcastReceiver = object: BroadcastReceiver(){
        override fun onReceive(
            context: Context,
            intent: Intent
        ){
            // check intent type
            if(intent.action != NfcAdapter.ACTION_ADAPTER_STATE_CHANGED) return

            // ignore if state is not final
            val state = intent.getIntExtra(
                NfcAdapter.EXTRA_ADAPTER_STATE,
                NfcAdapter.STATE_OFF
            )
            if(!arrayOf(
                NfcAdapter.STATE_ON,
                NfcAdapter.STATE_OFF
            ).contains(state)) return

            // update ui
            updateNfcEnabledState()
        }
    }
}
