package de.drmaxnix.mfcbalance.screen

import android.content.Intent
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.LazyListScope
import androidx.compose.material3.AlertDialogDefaults
import androidx.compose.material3.BasicAlertDialog
import androidx.compose.material3.ExperimentalMaterial3Api
import androidx.compose.material3.LocalContentColor
import androidx.compose.material3.LocalMinimumInteractiveComponentSize
import androidx.compose.material3.LocalTextStyle
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Scaffold
import androidx.compose.material3.Surface
import androidx.compose.material3.Text
import androidx.compose.material3.TextButton
import androidx.compose.runtime.Composable
import androidx.compose.runtime.CompositionLocalProvider
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.layout.Layout
import androidx.compose.ui.layout.Placeable
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.tooling.preview.Preview
import androidx.compose.ui.unit.Dp
import androidx.compose.ui.unit.dp
import androidx.compose.ui.util.fastForEach
import androidx.compose.ui.util.fastForEachIndexed
import androidx.core.net.toUri
import com.mikepenz.aboutlibraries.ui.compose.m3.LibrariesContainer
import de.drmaxnix.mfcbalance.ui.theme.MFCBalanceTheme
import me.zhanghai.compose.preference.Preference
import me.zhanghai.compose.preference.ProvidePreferenceLocals
import me.zhanghai.compose.preference.ProvidePreferenceTheme
import me.zhanghai.compose.preference.listPreference
import me.zhanghai.compose.preference.preference
import me.zhanghai.compose.preference.preferenceCategory
import me.zhanghai.compose.preference.preferenceTheme
import me.zhanghai.compose.preference.switchPreference
import me.zhanghai.compose.preference.textFieldPreference
import java.util.Locale
import kotlin.math.max

@Composable
fun Settings(){
    ProvidePreferenceLocals {
        ProvidePreferenceTheme(
            theme = preferenceTheme(
                categoryColor = MaterialTheme.colorScheme.primary
            )
        ) {
            val context = LocalContext.current
            LazyColumn(modifier = Modifier.fillMaxSize()) {
                preferenceCategory(
                    key = "cat_mfccard",
                    title = { Text(text = "Card Reader") },
                    modifier = Modifier.fillMaxWidth(),
                )
                textFieldPreference(
                    key = "mfccard_sector",
                    defaultValue = "0",
                    title = { Text(text = "Sector Number") },
                    textToValue = {
                        if(!Regex("[0-9]+").matches(it)) null
                        else {
                            val num = it.toInt()
                            if(num < 0 || num >= 40) null
                            else num
                        }
                    },
                    valueToText = { it.toString() },
                    summary = { Text(text = "Sector $it") },
                )
                textFieldPreference(
                    key = "mfccard_key",
                    defaultValue = "ffffffffffff",
                    title = { Text(text = "Sector Key") },
                    textToValue = {
                        if(!Regex("[0-9a-fA-F]{12}").matches(it)) null
                        else it.lowercase(Locale.ROOT)
                    },
                    valueToText = { it },
                    summary = { Text(text = it) },
                )
                textFieldPreference(
                    key = "mfccard_block_in_sector",
                    defaultValue = "0",
                    title = { Text(text = "Block Number in Sector") },
                    textToValue = {
                        if(!Regex("[0-9]+").matches(it)) null
                        else {
                            val num = it.toInt()
                            if(num < 0 || num >= 16) null
                            else num
                        }
                    },
                    valueToText = { it.toString() },
                    summary = { Text(text = "Block $it") },
                )
                textFieldPreference(
                    key = "mfccard_byte_offset_in_block",
                    defaultValue = "0",
                    title = { Text(text = "Byte Offset in Block") },
                    textToValue = {
                        if(!Regex("[0-9]+").matches(it)) null
                        else {
                            val num = it.toInt()
                            if(num < 0 || num >= 16) null
                            else num
                        }
                    },
                    valueToText = { it.toString() },
                    summary = { Text(text = "Byte $it") },
                )
                listPreference(
                    key = "mfccard_number_size",
                    defaultValue = 16,
                    values = listOf(8, 16, 24, 32),
                    valueToText = {
                        val byteNum = it / 8
                        AnnotatedString("$byteNum Byte" + if(byteNum != 1) "s" else "") },
                    title = { Text(text = "Number Size") },
                    summary = {
                        val byteNum = it / 8
                        Text(text = "$byteNum Byte" + if(byteNum != 1) "s" else "")
                    },
                )
                listPreference(
                    key = "mfccard_byteorder",
                    defaultValue = "be",
                    values = listOf("be", "le"),
                    valueToText = { AnnotatedString(if(it == "be") "Big-Endian" else "Little-Endian") },
                    title = { Text(text = "Byte-Order") },
                    summary = { Text(text = if(it == "be") "Big-Endian" else "Little-Endian") },
                )

                preferenceCategory(
                    key = "cat_balancedisplay",
                    title = { Text(text = "Balance Display") },
                    modifier = Modifier.fillMaxWidth(),
                )
                listPreference(
                    key = "balancedisplay_divisor",
                    defaultValue = 100,
                    values = listOf(1, 10, 100, 1000),
                    valueToText = { AnnotatedString(
                        when(it){
                            1 -> "No Decimal Point (eg. 27)"
                            10 -> "Tenths (eg. 27.4)"
                            100 -> "Cents (eg. 27.43)"
                            1000 -> "Thousandths (eg. 27.435)"
                            else -> ""
                        }
                    ) },
                    title = { Text(text = "Decimal Point Position") },
                    summary = { Text(
                        when(it){
                            1 -> "No Decimal Point"
                            10 -> "Tenths"
                            100 -> "Cents"
                            1000 -> "Thousandths"
                            else -> ""
                        }
                    ) },
                )
                listPreference(
                    key = "balancedisplay_currency_symbol",
                    defaultValue = "$",
                    values = listOf("$", "€", "£", "¥", "none"),
                    title = { Text(text = "Currency Symbol") },
                    summary = { Text(text = it) },
                )

                preferenceCategory(
                    key = "cat_debugging",
                    title = { Text(text = "Debugging") },
                    modifier = Modifier.fillMaxWidth(),
                )
                switchPreference(
                    key = "debugging_enabled",
                    defaultValue = false,
                    title = { Text(text = "Debug Mode") },
                    summary = { Text(text = "Show debug information") },
                )

                preferenceCategory(
                    key = "cat_about",
                    title = { Text(text = "About") },
                    modifier = Modifier.fillMaxWidth(),
                )
                preference(
                    key = "about_website",
                    title = { Text(text = "MFC Balance Website") },
                    summary = { Text(text = "Click to view the project's website") },
                    onClick = {
                        context.startActivity(Intent(Intent.ACTION_VIEW, "https://mfcbalance.drmaxnix.de".toUri()))
                    }
                )
                licenseView(
                    key = "about_libs",
                    title = { Text(text = "Used Libraries") },
                    summary = { Text(text = "Click to view licenses") },
                )
            }
        }
    }
}

inline fun LazyListScope.licenseView(
    key: String,
    crossinline title: @Composable () -> Unit,
    modifier: Modifier = Modifier.fillMaxWidth(),
    crossinline enabled: () -> Boolean = { true },
    noinline icon: @Composable (() -> Unit)? = null,
    noinline summary: @Composable (() -> Unit)? = null,
){
    item(key = key, contentType = "LicenseView") {
        LicenseView(
            title = { title() },
            modifier = modifier,
            enabled = enabled(),
            icon = icon?.let { { it() } },
            summary = summary?.let { { it() } },
        )
    }
}

@OptIn(ExperimentalMaterial3Api::class)
@Composable
fun LicenseView(
    title: @Composable () -> Unit,
    modifier: Modifier = Modifier,
    enabled: Boolean = true,
    icon: @Composable (() -> Unit)? = null,
    summary: @Composable (() -> Unit)? = null,
){
    var open by rememberSaveable { mutableStateOf(false) }
    if(open){
        BasicAlertDialog(
            onDismissRequest = { open = false },
            modifier = modifier,
        ){
            Surface(
                modifier = Modifier.fillMaxWidth(),
                shape = AlertDialogDefaults.shape,
                color = AlertDialogDefaults.containerColor,
                tonalElevation = AlertDialogDefaults.TonalElevation,
            ){
                Scaffold(
                    topBar = {
                        ProvideContentColorTextStyle(
                            contentColor = AlertDialogDefaults.titleContentColor,
                            textStyle = MaterialTheme.typography.headlineMedium,
                        ) {
                            Box(
                                modifier =
                                    Modifier
                                        .fillMaxWidth()
                                        .padding(
                                            start = 24.dp,
                                            top = 24.dp,
                                            end = 24.dp,
                                            bottom = 16.dp
                                        )
                            ) {
                                title()
                            }
                        }
                    },
                    content = {
                        padding -> Box(modifier = Modifier.padding(padding)){
                            LibrariesContainer(
                                Modifier.fillMaxSize()
                            )
                        }
                    },
                    bottomBar = {
                        ProvideContentColorTextStyle(
                            contentColor = MaterialTheme.colorScheme.primary,
                            textStyle = MaterialTheme.typography.labelLarge,
                        ) {
                            Box(
                                modifier =
                                    Modifier
                                        .fillMaxWidth()
                                        .padding(
                                            start = 24.dp,
                                            top = 16.dp,
                                            end = 24.dp,
                                            bottom = 24.dp
                                        ),
                                contentAlignment = Alignment.CenterEnd,
                            ) {
                                AlertDialogFlowRow(
                                    mainAxisSpacing = 8.dp,
                                    crossAxisSpacing = 12.dp
                                ) {
                                    CompositionLocalProvider(
                                        LocalMinimumInteractiveComponentSize provides Dp.Unspecified,
                                        content = {
                                            TextButton(onClick = { open = false }) {
                                                Text(text = "Close")
                                            }
                                        },
                                    )
                                }
                            }
                        }
                    },
                )
            }
        }
    }

    Preference(
        title = title,
        modifier = modifier,
        enabled = enabled,
        icon = icon,
        summary = summary,
    ) {
        open = true
    }
}

// Copied from androidx.compose.material3.ProvideContentColorTextStyle .
@Composable
private fun ProvideContentColorTextStyle(
    contentColor: Color,
    textStyle: TextStyle,
    content: @Composable () -> Unit,
) {
    val mergedStyle = LocalTextStyle.current.merge(textStyle)
    CompositionLocalProvider(
        LocalContentColor provides contentColor,
        LocalTextStyle provides mergedStyle,
        content = content,
    )
}

// Copied from androidx.compose.material3.AlertDialogFlowRow .
@Composable
private fun AlertDialogFlowRow(
    mainAxisSpacing: Dp,
    crossAxisSpacing: Dp,
    content: @Composable () -> Unit,
) {
    Layout(content) { measurables, constraints ->
        val sequences = mutableListOf<List<Placeable>>()
        val crossAxisSizes = mutableListOf<Int>()
        val crossAxisPositions = mutableListOf<Int>()

        var mainAxisSpace = 0
        var crossAxisSpace = 0

        val currentSequence = mutableListOf<Placeable>()
        var currentMainAxisSize = 0
        var currentCrossAxisSize = 0

        // Return whether the placeable can be added to the current sequence.
        fun canAddToCurrentSequence(placeable: Placeable) =
            currentSequence.isEmpty() ||
                    currentMainAxisSize + mainAxisSpacing.roundToPx() + placeable.width <=
                    constraints.maxWidth

        // Store current sequence information and start a new sequence.
        fun startNewSequence() {
            if (sequences.isNotEmpty()) {
                crossAxisSpace += crossAxisSpacing.roundToPx()
            }
            // Ensures that confirming actions appear above dismissive actions.
            @Suppress("ListIterator") sequences.add(0, currentSequence.toList())
            crossAxisSizes += currentCrossAxisSize
            crossAxisPositions += crossAxisSpace

            crossAxisSpace += currentCrossAxisSize
            mainAxisSpace = max(mainAxisSpace, currentMainAxisSize)

            currentSequence.clear()
            currentMainAxisSize = 0
            currentCrossAxisSize = 0
        }

        measurables.fastForEach { measurable ->
            // Ask the child for its preferred size.
            val placeable = measurable.measure(constraints)

            // Start a new sequence if there is not enough space.
            if (!canAddToCurrentSequence(placeable)) startNewSequence()

            // Add the child to the current sequence.
            if (currentSequence.isNotEmpty()) {
                currentMainAxisSize += mainAxisSpacing.roundToPx()
            }
            currentSequence.add(placeable)
            currentMainAxisSize += placeable.width
            currentCrossAxisSize = max(currentCrossAxisSize, placeable.height)
        }

        if (currentSequence.isNotEmpty()) startNewSequence()

        val mainAxisLayoutSize = max(mainAxisSpace, constraints.minWidth)

        val crossAxisLayoutSize = max(crossAxisSpace, constraints.minHeight)

        layout(mainAxisLayoutSize, crossAxisLayoutSize) {
            sequences.fastForEachIndexed { i, placeables ->
                val childrenMainAxisSizes =
                    IntArray(placeables.size) { j ->
                        placeables[j].width +
                                if (j < placeables.lastIndex) mainAxisSpacing.roundToPx() else 0
                    }
                val arrangement = Arrangement.End
                val mainAxisPositions = IntArray(childrenMainAxisSizes.size)
                with(arrangement) {
                    arrange(
                        mainAxisLayoutSize,
                        childrenMainAxisSizes,
                        layoutDirection,
                        mainAxisPositions,
                    )
                }
                placeables.fastForEachIndexed { j, placeable ->
                    placeable.place(x = mainAxisPositions[j], y = crossAxisPositions[i])
                }
            }
        }
    }
}

@Preview(showSystemUi = true)
@Composable
fun SettingsPreview(){
    MFCBalanceTheme {
        Settings()
    }
}
