package de.drmaxnix.mfcbalance.util

import android.nfc.Tag
import android.nfc.tech.MifareClassic
import java.io.IOException
import kotlin.math.ceil
import kotlin.math.pow
import kotlin.toUByte

class MfcReader(
    private val tag: Tag
){
    private var instructions: String = ""

    fun getNumber(
        sector: Int,
        keyBytes: ByteArray,
        blockInSector: Int,
        byteOffset: Int,
        numberSize: Int,
        byteorder: String
    ): UInt? {
        // GET AS MFC TAG //
        DebugLog.addLine(level = 0, line = "Getting MifareClassic tag object")
        val mfcTag = MifareClassic.get(tag)
        if(mfcTag == null){
            DebugLog.addLine(level = 1, line = "Object is null; aborting")
            instructions = "Incompatible Tag"
            return null
        }
        DebugLog.addLine(level = 1, line = "Success!")


        // VALIDATE SETTINGS VALUES //
        DebugLog.addLine(level = 0, line = "Validating settings")

        // sector
        val highestSector = mfcTag.sectorCount - 1
        DebugLog.addLine(level = 1, line = "Highest sector index: $highestSector")
        if(sector > highestSector){
            DebugLog.addLine(level = 2, line = "Sector ($sector) out of range; aborting")
            instructions = "Sector ($sector) out of Range (0 - $highestSector)"
            return null
        }

        // block
        val highestBlockInSector = mfcTag.getBlockCountInSector(sector) - 1
        DebugLog.addLine(level = 1, line = "Highest block index in sector: $highestBlockInSector")
        if(blockInSector > highestBlockInSector){
            DebugLog.addLine(level = 2, line = "Block ($blockInSector) out of range; aborting")
            instructions = "Block ($blockInSector) out of Range (0 - $highestBlockInSector)"
            return null
        }

        // number size
        val highestByteOffset = (16 - (numberSize / 8))
        DebugLog.addLine(level = 1, line = "Highest possible byte offset: $highestByteOffset")
        if(byteOffset > highestByteOffset){
            DebugLog.addLine(level = 2, line = "Byte offset ($byteOffset) out of range; aborting")
            instructions = "Byte Offset ($byteOffset) out of Range (0 - $highestByteOffset)"
            return null
        }


        // READ BLOCK FROM CARD //
        DebugLog.addLine(level = 0, line = "Reading data")

        // connect
        try {
            mfcTag.connect()
        } catch (_: IOException){
            DebugLog.addLine(level = 1, line = "Failed to connect; aborting")
            instructions = "Connection Error"
            return null
        }
        DebugLog.addLine(level = 1, line = "Connected")

        // authenticate
        var hasAuth = false
        for(oneKeySlot in arrayOf("A", "B")){
            DebugLog.addLine(level = 1, line = "Trying auth on key slot $oneKeySlot")
            try {
                val success =
                    if(oneKeySlot == "A") mfcTag.authenticateSectorWithKeyA(sector, keyBytes)
                    else mfcTag.authenticateSectorWithKeyB(sector, keyBytes)
                if(success){
                    DebugLog.addLine(level = 2, line = "Success!")
                    hasAuth = true
                    break
                }
                DebugLog.addLine(level = 2, line = "Failure (returned false)")

            } catch(_: IOException){
                DebugLog.addLine(level = 2, line = "Failure (IOException)")
            }
        }
        if(!hasAuth){
            DebugLog.addLine(level = 1, line = "Failed to authenticate; aborting")
            instructions = "Authentication Error"
            return null
        }
        DebugLog.addLine(level = 1, line = "Authenticated")

        // read block
        val sectorBlockOffset = mfcTag.sectorToBlock(sector)
        val blockIndex = sectorBlockOffset + blockInSector
        DebugLog.addLine(level = 1, line = "Reading block at index $blockIndex")
        val blockData: ByteArray
        try {
            blockData = mfcTag.readBlock(blockIndex)
        } catch(_: IOException){
            DebugLog.addLine(level = 2, line = "Reading failed; aborting")
            instructions = "Read Error"
            return null
        }
        DebugLog.addLine(level = 2, line = "Success!")
        DebugLog.addLine(level = 2, line = "Block data: ${byteArrayToHexString(blockData)}")

        // disconnect
        try {
            mfcTag.close()
        } catch(_: IOException){
            DebugLog.addLine(level = 1, line = "Failed to disconnect; aborting")
            instructions = "Connection Error"
            return null
        }
        DebugLog.addLine(level = 1, line = "Disconnected")


        // EXTRACT NUMBER FROM BLOCK DATA //
        return numberExtract(
            blockData = blockData,
            byteOffset = byteOffset,
            numberSize = numberSize,
            byteorder = byteorder
        )
    }

    fun getInstructions(): String {
        return instructions
    }

    companion object {
        fun hexStringToByteArray(string: String): ByteArray {
            val arraySize = ceil(string.length.toDouble() / 2.0f).toInt()
            val bytes = ByteArray(arraySize)
            for(q in 0..<arraySize){
                val byteString = string.substring(startIndex = q * 2, endIndex = (q * 2) + 2)
                val byte = Integer.parseInt(byteString, 16).toByte()
                bytes.set(index = q, byte)
            }
            return bytes
        }

        fun byteArrayToHexString(array: ByteArray): String {
            var string = ""
            array.forEach {
                string += it.toUByte().toInt().toString(16).padStart(2, '0')
            }
            return string
        }

        fun numberExtract(
            blockData: ByteArray,
            byteOffset: Int,
            numberSize: Int,
            byteorder: String
        ): UInt {
            DebugLog.addLine(level = 0, line = "Extracting number")

            // get number bytes
            val fromByte = byteOffset
            val toByte = fromByte + (numberSize / 8)
            val numberBytes = blockData.copyOfRange(fromByte, toByte)
            DebugLog.addLine(level = 1, line = "Number bytes ($fromByte - $toByte): ${byteArrayToHexString(numberBytes)}")

            // maybe reverse to match byte order
            if(byteorder == "le") numberBytes.reverse()
            DebugLog.addLine(level = 1, line = "Byte order corrected ($byteorder): ${byteArrayToHexString(numberBytes)}")

            // build unsigned integer from byte array
            DebugLog.addLine(level = 1, line = "Assembling number from bytes")
            var number = 0u
            for(q in numberBytes.lastIndex downTo 0){
                val weight = 256f.pow(q).toUInt()
                val amount = numberBytes[(numberBytes.lastIndex - q)].toUByte().toUInt()
                val positionValue = weight * amount
                DebugLog.addLine(level = 2, line = "Adding: $weight * $amount = $positionValue")
                number += positionValue
            }
            DebugLog.addLine(level = 2, line = "Number: $number")

            // done
            return number
        }
    }
}